"""
This module is a plugin that adds functionality for selecting data that has been acquired using the new
"Acquisition Sequencer" in the PWS acquisition software.
"""

from ._plugin import AcquisitionSequencerPlugin


